//###############################################################################
//#
//# File:         Account.java
//# RCS:          $Header: /u/ejk/var/cvs/courses/cs161/login-solution/Account.java,v 1.2 2005/12/02 13:11:18 ejk Exp $
//# Description:  A single-server IMAP Client encapsulated in a thread
//# Author:       Eric Koskinen
//# Created:      Wed Sep 14 12:48:49 2005
//# Modified:     Wed Sep 14 12:48:49 2005 (Eric Koskinen) ejk@cs.brown.edu
//# Language:     Java
//# Package:      N/A
//# Status:       Experimental (Do Not Distribute)
//#
//# (C/) Copyright 2005, Eric J Koskinen, all rights reserved.
//#
//###############################################################################

import java.net.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.nio.charset.Charset;
import java.util.*;

public class Account extends Thread {

    private String host;
    private String port;
    private SocketChannel sc;
    private Socket s;
    private InetSocketAddress sa;
    private ArrayList<String> commands;
    private int commandTag;
    private Charset charset = Charset.forName("UTF-8");
    private ByteBuffer out_buf;
    private ByteBuffer in_buf = ByteBuffer.allocate( 1024 );

    // ****************************************************************
    //
    // constructors
    //
    // ****************************************************************

    public Account(String n) { super(n); }
    public Account() { super(); }

    // ****************************************************************
    //
    // init: set up a socket to communicate with the given IMAP server
    //
    // ****************************************************************

    public void init(String h, String p) throws IOException {
        host = h;
        port = p;
	sc = SocketChannel.open();
	sc.configureBlocking( true );
	s  = sc.socket();
	sa = new InetSocketAddress( host, new Integer(port) );
	try {
	    sc.connect( sa );
	} catch (ConnectException e) {
	    System.out.println("["+this.getName()+"] Connect failed: "+
			     e.getMessage()+".");
	}
	    
	commands = new ArrayList<String>();
	commandTag = 0;
    }

    // ****************************************************************
    //
    // run: perform each command in the queue
    //
    // ****************************************************************

    public void run() {
	try {
	    if (!sc.isConnected())
		sc.finishConnect();
	    in_buf.limit(0);
	    // get the connection 
	    read_line();
	    // do commands
	    for (int i=0; i<commands.size(); i++) {
		String tag = String.format("A%03d", commandTag++);
		write_command(commands.get(i), tag);
		String response = read_response(tag);
		// handle BAD responses
		if (response.indexOf(tag+" BAD ")>=0) {
		    System.out.println("["+this.getName()+"] *** BAD happened *** closing");
		    break;
		}
		// deal with the response
		if(!handleCommand(tag, commands.get(i), response)) break;
	    }
	    // close up
	    sc.close();
	} catch (IOException e) {
	    System.out.println("["+this.getName()+"] IOException: "+
			     e.getMessage()+".");
	};
	System.out.println("["+this.getName()+"] DONE.");
    }

    // ****************************************************************
    //
    // write_command: (BLOCKING) writes a given IMAP command to the 
    //   server, prefixed with the given tag
    //
    // ****************************************************************

    private void write_command(String cmd, String tag) {
	// create the cmd
	String line = String.format("%s %s\r\n", tag, cmd);
	System.out.print("["+this.getName()+"] SEND >>> "+line);
	// write it out
	ByteBuffer out_buf = ByteBuffer.allocate( 1024 );
	out_buf.clear();
	for (int i = 0; i<line.length(); i++)
	    out_buf.put((byte)line.charAt(i));
	try {
	    out_buf.flip();
	    sc.write( out_buf );
	} catch (IOException e) {
	    System.out.println("["+this.getName()+"] IOException: "+
			     e.getMessage()+".");
	}
    }

    // ****************************************************************
    //
    // read_response: (BLOCKING) reads one line at a time from the
    //   IMAP server until finding a line that's prefixed with given tag
    //
    // ****************************************************************

    private String read_response(String tag) {
	String response = new String("");
	String delta    = new String("");
	do {
	    //System.out.println("looking for "+tag+" ...");
	    delta = read_line();
	    response = response.concat(delta);
	} while ( !delta.regionMatches(0,tag,0,tag.length()) );
	return response;
    }

    // ****************************************************************
    //
    // read_line: (BLOCKING) reads from the SocketChannel until finding
    //   a newline character.
    //
    // ****************************************************************

    private String read_line() {
	//System.out.println("read_line "+in_buf.toString());
	StringBuffer str = new StringBuffer( 1024 );
	try {
	    do {
		// copy
		while(in_buf.hasRemaining()) {
		    Character c = new Character((char)in_buf.get());
		    str.append( c );
		    if (c.compareTo(new Character('\n'))==0) { 
			//System.out.print("read_line = "+str.toString());
			System.out.print("["+this.getName()+"] RECV <<< "+str.toString());
			return str.toString();
		    }
		}
		// get some data
		in_buf.clear();
		sc.read( in_buf );
		in_buf.flip();
	    } while (true);
	} catch (IOException e) {
	    System.out.println("["+this.getName()+"] IOException: "+
			     e.getMessage()+".");
	}
	return str.toString();
    }

    // ****************************************************************
    //
    // queueCommands: stores the provided commands in this object
    //
    // ****************************************************************

    public void queueCommands(List cmds) {
	for (int i=0;i<cmds.size(); i++) {
	    commands.add((String)cmds.get(i));
	}
    }

    // ****************************************************************
    //
    // handleCommand: perform command-specific tasks
    //
    //   Return false if we should disconnect, true to continue.
    //
    // ****************************************************************

    private boolean handleCommand(String tag, String cmd, String response) {
	
	// +++ LOGIN +++
	if (cmd.compareTo("NOOP")==0) {

	    // do nothing

	} else if (cmd.compareTo("CAPABILITY")==0) {

	    // unimplemented

	} else if (cmd.regionMatches(0,"LOGIN ",0,6)) {

	    if (response.indexOf(tag+" NO ")>=0) {
		System.out.println("["+this.getName()+"] *** LOGIN FAILED ***");
		return false;
	    } else {
		// unimplemented
	    }

	} else if (cmd.compareTo("LOGOUT")==0) {

	    return false;

	} else {

	    System.out.println("["+this.getName()+"] *** Unrecognized command: "+cmd);

	}

	return true;
    }

}
